import copy
import numpy as np
import pandas as pd
import xgboost as xgb
from sklearn.model_selection import train_test_split


class XGBRegressor:

    def __init__(self, **params):
        """
        XG Boost wrapper
        @param params: Model parameters
        """
        default = {'verbosity': 0}
        for k, v in default.items():
            if k not in params.keys():
                params[k] = v
        self.default = default
        self.params = params
        self.set_params(**self.params)
        self.model = None
        self.callbacks = None
        self.trained = False
        self._estimator_type = 'regressor'
        self.binary = False

    @staticmethod
    def convert_to_d_matrix(x, y=None):
        # Convert input
        assert type(x) in [pd.DataFrame, pd.Series, np.ndarray], 'Unsupported data input format'
        if isinstance(x, np.ndarray) and len(x.shape) == 0:
            x = x.reshape((-1, 1))

        if y is None:
            return xgb.DMatrix(x)

        else:
            assert type(y) in [pd.Series, np.ndarray], 'Unsupported data label format'
            return xgb.DMatrix(x, label=y)

    def fit(self, x, y):
        # Split & Convert data
        train_x, test_x, train_y, test_y = train_test_split(x, y, test_size=0.1)
        d_train = self.convert_to_d_matrix(train_x, train_y)
        d_test = self.convert_to_d_matrix(test_x, test_y)

        # Model training
        self.model = xgb.train(self.params,
                               d_train,
                               evals=[(d_test, 'validation'), (d_train, 'train')],
                               verbose_eval=False,
                               callbacks=[self.callbacks] if self.callbacks is not None else None,
                               early_stopping_rounds=100,
                               )
        self.trained = True

    def predict(self, x):
        # todo check input data
        assert self.trained is True, 'Model not yet trained'
        d_predict = self.convert_to_d_matrix(x)
        return self.model.predict(d_predict)

    def set_params(self, **params):
        for k, v in self.default.items():
            if k not in params.keys():
                params[k] = v
        if 'callbacks' in params.keys():
            self.callbacks = params['callbacks']
            params.pop('callbacks')
        self.params = params
        return self

    def get_params(self, **args):
        params = copy.copy(self.params)
        if 'deep' in args:
            return params
        if self.callbacks is not None:
            params['callbacks'] = self.callbacks
        return params



import json
import unittest
import numpy as np
import pandas as pd
from Amplo.AutoML import DriftDetector


def draw(n):
    return pd.DataFrame({
        'norm': np.random.normal(0.23, 0.53, n),
        'uniform': np.random.uniform(0, 100, n),
        'exponential': np.random.exponential(0.4, n),
        'gamma': np.random.gamma(0.3, 0.9, n),
        'beta': np.random.beta(0.2, 0.4, n),
    })


class TestDriftDetector(unittest.TestCase):

    def test_distribution_fits(self):
        # Setup
        dists = ["norm", "uniform", "expon", "gamma", "beta"]
        ref = draw(500)
        test = ref.iloc[np.random.permutation(len(ref))[:10]]
        drift = DriftDetector(num_cols=ref.keys())
        drift.fit(ref)

        # Checks
        assert len(drift.check(test)) == 0, "Test data found inconsistent"
        assert len(drift.check(ref.max() + 1)) == len(dists), "Maxima not detected"
        assert len(drift.check(ref.min() - 1)) == len(dists), "Minima not detected"

    def test_categorical(self):
        df = pd.DataFrame({'a': ['a', 'b', 'c', 'd', 'a', 'b', 'a', 'b', 'c', 'a']})
        drift = DriftDetector(cat_cols=['a'])
        drift.fit(df)
        assert 'a' in drift.bins, "Column 'a' rejected."
        assert drift.bins['a'] == {'a': 4, 'b': 3, 'c': 2, 'd': 1}

    def test_add_bins(self):
        df = pd.DataFrame({'a': ['a', 'b', 'c', 'd', 'a', 'b', 'a', 'b', 'c', 'a']})
        yp = np.random.randint(0, 2, (100))
        drift = DriftDetector(cat_cols=['a'])
        drift.fit(df)

        # Test empty
        assert drift.add_bins({}, df)
        assert drift.add_output_bins((), yp)

        # Test actual adding
        new_bins = drift.add_bins(drift.bins, df)
        assert new_bins['a'] == {'a': 8, 'b': 6, 'c': 4, 'd': 2}

    def test_storable(self):
        df = pd.DataFrame({'a': ['a', 'b', 'c'], 'b': [0, 0.1, 0.2]})
        drift = DriftDetector(cat_cols=['a'], num_cols=['b'])
        drift.fit(df)
        json.dumps(drift.bins)
        json.dumps(drift.add_bins(drift.bins, df))
        pred = np.random.randint(0, 2, (100))
        old = drift.add_output_bins((), pred)
        drift.add_output_bins(old, pred)




import os
import shutil
import unittest
import pandas as pd
from sklearn.datasets import make_classification, make_regression
from Amplo import Pipeline
from Amplo.AutoML import Modeller


class TestPipeline(unittest.TestCase):

    def test_main_predictors(self):
        if os.path.exists('AutoML'):
            shutil.rmtree('AutoML')
        for mode in ['classification', 'regression']:
            if mode == 'classification':
                x, y = make_classification()
            else:
                x, y = make_regression()
            x, y = pd.DataFrame(x), pd.Series(y)
            pipeline = Pipeline(grid_search_iterations=1, grid_search_candidates=1, plot_eda=False)
            pipeline.fit(x, y)
            for samples in [100, 25001]:
                for model in Modeller(mode=mode, samples=samples).return_models():
                    print(model)
                    x_c, _ = pipeline.convert_data(x)
                    model.fit(x_c, y)
                    pipeline.bestModel = model
                    pipeline.predict(x)
                    assert isinstance(pipeline._main_predictors, dict), 'Main predictors not dictionary.'

            if os.path.exists('AutoML'):
                shutil.rmtree('AutoML')

    def test_no_dirs(self):
        if os.path.exists('AutoML'):
            shutil.rmtree('AutoML')
        pipeline = Pipeline(no_dirs=True)
        assert not os.path.exists('AutoML'), 'Directory created'

    def test_no_args(self):
        if os.path.exists('AutoML'):
            shutil.rmtree('AutoML')
        x, y = make_regression()
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(x, y)
        shutil.rmtree('AutoML')

    def test_mode_detector(self):
        x, y = make_regression()
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(x, y)
        assert pipeline.mode == 'regression'
        shutil.rmtree('AutoML')
        x, y = make_classification()
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(x, y)
        assert pipeline.mode == 'classification'
        shutil.rmtree('AutoML')

    def test_create_folders(self):
        if os.path.exists('AutoML'):
            shutil.rmtree('AutoML')

        x, y = make_classification()
        pipeline = Pipeline(grid_search_iterations=0)
        pipeline.fit(x, y)

        # Test Directories
        assert os.path.exists('AutoML')
        assert os.path.exists('AutoML/Data')
        assert os.path.exists('AutoML/Features')
        assert os.path.exists('AutoML/Production')
        assert os.path.exists('AutoML/Documentation')
        assert os.path.exists('AutoML/Results.csv')



import os
import pytest
import shutil
import unittest
import numpy as np
import pandas as pd
from Amplo.AutoML import IntervalAnalyser


def createDataFrames(n_samples, n_features):
    dim = (int(n_samples / 2), n_features)
    columns = [f'Feature_{i}' for i in range(n_features)]
    df1 = pd.DataFrame(
        columns=columns,
        data=np.vstack((np.random.normal(0, 1, dim), np.random.normal(100, 1, dim))))
    df2 = pd.DataFrame(
        columns=columns,
        data=np.vstack((np.random.normal(0, 1, dim), np.random.normal(-100, 1, dim))))
    return df1, df2


class TestIntervalAnalyser(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        # Settings
        cls.n_samples = 50
        cls.n_features = 25

        # Create classes
        if not os.path.exists('IA/Class_1'):
            os.makedirs('IA/Class_1')
            os.makedirs('IA/Class_2')

        for i in range(140):
            # Create dataframes
            df1, df2 = createDataFrames(cls.n_samples, cls.n_features)
            df1.to_csv(f'IA/Class_1/Log_{i}.csv', index=False)
            df2.to_csv(f'IA/Class_2/Log_{i}.csv', index=False)

    @classmethod
    def tearDownClass(cls) -> None:
        if os.path.exists('IA'):
            shutil.rmtree('IA')
        if os.path.exists('tests/Unit/AutoML/IA'):
            shutil.rmtree('tests/Unit/AutoML/IA')

    def test_all(self):
        ia = IntervalAnalyser(folder='IA', min_length=10)
        df = ia.fit_transform()

        # Attribute tests
        assert ia.n_folders == 2
        assert ia.n_files == 280

        # Functional tests
        for i in range(ia.n_files):
            dist = ia._distributions[i].values
            assert all(v < 0.8 for v in dist[:int(len(dist) / 2)]), 'Noise with high percentage of neighbors'
            assert all(v > 0.8 for v in dist[int(len(dist) / 2):]), 'Information with low percentage of neighbors'
        assert len(df) == int(ia.samples / 2), 'Incorrect number of samples'
        assert df.index.get_level_values(0).nunique() == ia.n_files, 'Files skipped'







import copy
import warnings
import numpy as np
import pandas as pd
import lightgbm as lgb
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import train_test_split


class LGBMRegressor:

    def __init__(self, **params):
        """
        Light GBM wrapper
        @param params: Model parameters
        
        callbacks [list[str]]: Callbacks as a string, transformed in a function here
        """
        default = {'verbosity': -1, 'force_col_wise': True}
        if not 'objective' in params:
            default['objective'] = 'rmse'
        for k, v in default.items():
            if k not in params.keys():
                params[k] = v

        self.params = params
        self.default = default
        self.set_params(**self.params)
        self.model = None
        self.callbacks = []
        self.trained = False
        self._estimator_type = 'regressor'

    @staticmethod
    def convert_to_dataset(x, y=None):
        # Convert input
        assert type(x) in [pd.DataFrame, pd.Series, np.ndarray], 'Unsupported data input format'
        if isinstance(x, np.ndarray) and len(x.shape) == 0:
            x = x.reshape((-1, 1))

        if y is None:
            return lgb.Dataset(x)

        else:
            assert type(y) in [pd.Series, np.ndarray], 'Unsupported data label format'
            return lgb.Dataset(x, label=y)

    def fit(self, x, y):
        assert isinstance(x, np.ndarray) or isinstance(x, pd.DataFrame), 'X needs to be of type np.ndarray or ' \
                                                                         'pd.DataFrame'
        assert isinstance(y, np.ndarray) or isinstance(y, pd.Series), 'Y needs to be of type np.ndarray or pd.Series'
        if isinstance(x, pd.DataFrame):
            x = x.to_numpy()
        if isinstance(y, pd.Series):
            y = y.to_numpy()

        # Split & Convert data
        train_x, test_x, train_y, test_y = train_test_split(x, y, test_size=0.1, shuffle=True)
        d_train = self.convert_to_dataset(train_x, train_y)
        d_test = self.convert_to_dataset(test_x, test_y)

        # Model training
        self.model = lgb.train(self.params,
                               d_train,
                               valid_sets=[d_train, d_test],
                               feval=self.eval_function,
                               callbacks=[lgb.early_stopping(100, verbose=False)] + self._get_callbacks(),
                               )
        self.trained = True

    def predict(self, x):
        # todo check input data
        assert self.trained is True, 'Model not yet trained'
        return self.model.predict(x)

    def set_params(self, **params):
        for k, v in self.default.items():
            if k not in params.keys():
                params[k] = v
        if 'callbacks' in params.keys():
            self.callbacks = params['callbacks']
            params.pop('callbacks')
        self.params = params
        return self

    def _get_callbacks(self) -> list:
        # Output
        callbacks = []

        # Iterate through callbacks
        for callback in self.callbacks:
            # todo implement various callbacks - recognize string and add function
            raise NotImplementedError('Callback not implemented.')

        # Return
        return callbacks

    def get_params(self, **args):
        params = copy.copy(self.params)
        if 'deep' in args:
            return params
        if self.callbacks is not None:
            params['callbacks'] = self.callbacks
        return params

    @staticmethod
    def eval_function(prediction, d_train):
        target = d_train.get_label()
        weight = d_train.get_weight()
        # Return mean absolute error
        return "mean_absolute_error", mean_absolute_error(target, prediction, sample_weight=weight), True



import os
import copy
import faiss
import numpy as np
import pandas as pd
from Amplo.AutoML import DataProcesser
from Amplo.AutoML import FeatureProcesser


class IntervalAnalyser:

    def __init__(self,
                 folder: str = None,
                 norm: str = 'euclidean',
                 min_length: int = 1000,
                 n_neighbors: int = None,
                 n_trees: int = 10,
                 verbose: int = 1,
                 ):
        """
        Interval Analyser for Log file classification. Has two purposes:
        - Remove healthy data in longer, faulty logs
        - Remove redundant data in large datasets

        Uses Facebook's FAISS for K-Nearest Neighbors approximation.

        ** IMPORTANT **
        To use this interval analyser, make sure that your logs are located in a folder of their label, with one parent
        folder with all labels, e.g.:
        +-- Parent Folder
        |   +-- Label_1
        |       +-- Log_1.*
        |       +-- Log_2.*
        |   +-- Label_2
        |       +-- Log_3.*

        Parameters
        ----------
        folder [str]:       Parent folder of labels
        index_col [str]:    For reading the log files
        min_length [int]:   Minimum length to cut off, everything shorter is left untouched
        norm [str]:         Optimization metric for K-Nearest Neighbors
        n_neighbors [int]:  Quantity of neighbors, default to 3 * log length
        n_trees [int]:      Quantity of trees
        """
        # Parameters
        self.folder = folder + '/' if len(folder) == 0 or folder[-1] != '/' else folder
        self.min_length = min_length
        self.norm = norm
        self.n_trees = n_trees
        self.n_neighbors = n_neighbors
        self.verbose = verbose

        # Initializers
        self.keys = None
        self.n_keys = None
        self.samples = None
        self.avg_samples = None
        self.n_files = 0
        self.n_folders = 0
        self._mins = pd.DataFrame(index=[0])
        self._maxs = pd.DataFrame(index=[0])
        self._labels = None
        self._engine = None
        self._distributions = None
        self._str_labels = None

        # Test
        assert norm in ['euclidean', 'manhattan', 'angular', 'hamming', 'dot']

    def fit_transform(self) -> pd.DataFrame:
        """
        Function that runs the K-Nearest Neighbors and returns a NumPy array with the sensitivities.

        Returns
        -------
        np.array: Estimation of strength of correlation
        """
        # Get data
        df, labels = self._parse_data()

        # Set up Annoy Engine (only now that n_keys is known)
        self._engine = self._build_engine(df)

        # Make distribution
        self._distributions = self._make_distribution(df, labels)

        # Return new dataset
        return self._create_dataset(df, labels)

    def _create_dataset(self, df: pd.DataFrame, labels: pd.Series) -> pd.DataFrame:
        """
        This function selects samples given the calculated distributions. It only removes samples from logs which are
        longer (> min_length), and only the samples with lower in-class neighbors.

        One could come up with a fancier logic, using the total dataset samples, the class-balance & sample redundancy.

        parameters
        ----------
        """
        # Verbose
        if self.verbose > 0:
            print('[AutoML] Creating filtered dataset')

        # Get in-class means and number of samples for each label
        # label_means = self._get_label_means(labels, self._distributions)
        # label_samples = labels.value_counts()

        # Initialize
        ind_remove = []

        # Iterate through labels and see if we should remove values
        for i in range(self.n_files):

            # Check length and continue if short
            if sum(df.index.get_level_values(0) == i) < self.min_length:
                continue

            # Check distribution and find cut-off
            dist = self._distributions[i]
            ind_remove.extend([(i, j) for j in np.where(dist < dist.mean())[0]])

            # Verbose
            if len(ind_remove) > 0 and self.verbose > 1:
                print(f'[AutoML] Removing {len(ind_remove)} samples from {os.listdir(self.folder)[labels[(i, 0)]]}')

        # Remove samples from df
        df = df.drop(ind_remove, axis=0)

        # Return original scales
        df = df * (self._maxs[df.keys()] - self._mins[df.keys()]) + self._mins[df.keys()]

        # Add labels
        df['labels'] = self._str_labels

        return df

    def _parse_data(self):
        """
        Reads all files and sets a multi index

        Returns
        -------
        pd.DataFrame: all log files
        """
        # Verbose
        if self.verbose > 0:
            print('[AutoML] Parsing data for interval analyser.')

        # Result init
        dfs = []

        # Loop through files
        for folder in os.listdir(self.folder):
            for file in os.listdir(self.folder + folder):

                # Verbose
                if self.verbose > 1:
                    print(f"[AutoML] {self.folder}{folder}/{file}")

                # Read df
                df = self._read(f'{self.folder}{folder}/{file}')

                # Set label
                df['labels'] = folder

                # Set second index
                df = df.set_index(pd.MultiIndex.from_product([[self.n_files], df.index.values], names=['log', 'index']))

                # Add to list
                dfs.append(df)

                # Increment
                self.n_files += 1
            self.n_folders += 1

        # Concatenate dataframes
        dfs = pd.concat(dfs)

        # Extract string labels
        self._str_labels = copy.deepcopy(dfs['labels'])

        # Clean data
        dp = DataProcesser(missing_values='zero', target='labels')
        dfs = dp.fit_transform(dfs)

        # Remove datetime columns
        if len(dp.date_cols) != 0:
            dfs = dfs.drop(dp.date_cols, axis=1)

        # Remove classes
        labels = dfs['labels']
        dfs = dfs.drop('labels', axis=1)

        # Select features -- bit hacky to avoid memory
        fp = FeatureProcesser(extract_features=False, mode='classification')
        fp.x, fp.y = dfs, labels
        features, _ = fp._sel_gini_impurity()
        dfs = dfs[features]

        # Normalize
        self._mins, self._maxs = dfs.min(), dfs.max()
        dfs = (dfs - dfs.min()) / (dfs.max() - dfs.min())

        # Set sizes
        self.samples = len(dfs)
        self.n_keys = len(dfs.keys())
        if self.n_neighbors is None:
            self.n_neighbors = min(int(3 * self.samples / self.n_files), 5000)

        # Return
        return dfs, labels

    def _build_engine(self, df):
        """
        Builds the ANNOY engine.
        """
        # Create engine
        if self.verbose > 0:
            print('[AutoML] Building interval analyser engine.')
        engine = faiss.IndexFlatL2(self.n_keys)

        # Add the data to ANNOY
        engine.add(np.ascontiguousarray(df.values))

        return engine

    def _read(self, path: str) -> pd.DataFrame:
        """
        Wrapper for various read functions
        """
        f_ext = path[path.rfind('.'):]
        if f_ext == '.csv':
            return pd.read_csv(path)
        elif f_ext == '.json':
            return pd.read_json(path)
        elif f_ext == '.xml':
            return pd.read_xml(path)
        elif f_ext == '.feather':
            return pd.read_feather(path)
        elif f_ext == '.parquet':
            return pd.read_parquet(path)
        elif f_ext == '.stata':
            return pd.read_stata(path)
        elif f_ext == '.pickle':
            return pd.read_pickle(path)
        else:
            raise NotImplementedError('File format not supported.')

    def _make_distribution(self, df: pd.DataFrame, labels: pd.Series) -> pd.Series:
        """
        Given a build K-Nearest Neighbors, returns the label distribution
        """
        if self.verbose > 0:
            print('[AutoML] Calculating interval within-class distributions.')

        # Search nearest neighbors for all samples -- has to be iterative for large files -.-
        distribution = []
        for i, row in df.iterrows():
            _, neighbors = self._engine.search(np.ascontiguousarray(row.values.reshape((1, -1))), self.n_neighbors)
            distribution.append(sum(labels.iloc[neighbors.reshape(-1)] == labels.loc[i]) / self.n_neighbors)

        # Parse into list of lists
        return pd.Series(distribution, index=labels.index)

    @staticmethod
    def _get_label_means(labels, dists):
        # Init
        label_means = {k: [] for k in labels.unique()}

        # Append distributions
        for i, d in enumerate(dists):
            # Skip failed reads
            if i not in labels:
                continue

            # Extend distribution
            label_means[labels[(i, 0)]].extend(d)

        # Return means
        return {k: np.mean(v) for k, v in label_means.items()}



import os
import copy
import numpy as np
import pandas as pd
from tqdm import tqdm
import matplotlib
import matplotlib.pyplot as plt
from statsmodels.tsa.seasonal import STL
from statsmodels.graphics.tsaplots import plot_acf
from statsmodels.graphics.tsaplots import plot_pacf
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import RandomForestClassifier


class DataExplorer:

    def __init__(self, data,
                 y=None,
                 mode: str = 'regression',
                 plot_time_plots: bool = True,
                 plot_box_plots: bool = False,
                 plot_missing_values: bool = True,
                 plot_seasonality: bool = False,
                 plot_co_linearity: bool = True,
                 plot_differencing: bool = False,
                 plot_signal_correlations: bool = False,
                 plot_feature_importance: bool = True,
                 plot_scatter_plots: bool = False,
                 differ: int = 0,
                 pre_tag: str = '',
                 max_samples: int = 10000,
                 season_periods: int = 24,
                 lags: int = 60,
                 skip_completed: bool = True,
                 folder: str = '',
                 version: str = 'v0'):
        """
        Automated Exploratory Data Analysis. Covers binary classification and regression.
        It generates:
        - Missing Values Plot
        - Line Plots of all features
        - Box plots of all features
        - Co-linearity Plot
        - SHAP Values
        - Random Forest Feature Importance
        - Predictive Power Score

        Additionally fFor Regression:
        - Seasonality Plots
        - Differentiated Variance Plot
        - Auto Correlation Function Plot
        - Partial Auto Correlation Function Plot
        - Cross Correlation Function Plot
        - Scatter Plots

        All figures are organized per project, category and version with the following file structure:
        |-- 'Folder parameter'
        |   |-- EDA
        |   |   |-- Timeplots
        |   |   |   |-- v'i'
        |   |   |   |   |-- Feature_0.png
        |   |   |   |   |-- Feature_1.png
        ...

        Parameters
        ----------
        data pd.DataFrame: Input data, either Numpy or Pandas
        y pd.Series: Output / Target, either Numpy or Pandas
        mode str: classification or regression
        plot_time_plots bool: Boolean to create time plots or not
        plot_box_plots bool: Boolean to create box plots
        plot_missing_values bool: Boolean to create missing values plot
        plot_seasonality bool: Boolean to create seasonality plots
        plot_co_linearity bool: Boolean to create co-linearity plots
        plot_differencing bool: Boolean to create differencing plots
        plot_signal_correlations bool: Boolean to create signal correlation plots (ACF, PACF, CCF)
        plot_feature_importance bool: Boolean to create feature importance plots (SHAP, RF, PPS)
        plot_scatter_plots bool: Boolean to create scatter plots
        differ int: Differencing order for signal correlation plots
        pre_tag str: Optional, string to format file names
        max_samples int: Optional, under-sampling for large datasets
        season_periods int: Optional, season periods for seasonality plots
        lags int: Optional, maximum lags to take into account for signal plots
        skip_completed bool: Whether to create existing plots
        folder str: Main folder to dump files in
        version int: All graphs are put in a versioned folder
        """
        assert isinstance(data, np.ndarray) or isinstance(data, pd.DataFrame), 'Data should be Numpy or Pandas'
        if y is not None:
            assert isinstance(y, np.ndarray) or isinstance(y, pd.Series), 'Y should be Numpy or Pandas'
        if type(data) != pd.DataFrame:
            data = pd.DataFrame(data=data, columns=['Feature_{}'.format(i) for i in range(data.shape[1])])
        assert mode in ['classification', 'regression']

        # Running booleans
        self.plotTimePlots = plot_time_plots
        self.plotBoxPlots = plot_box_plots
        self.plotMissingValues = plot_missing_values
        self.plotSeasonality = plot_seasonality
        self.plotCoLinearity = plot_co_linearity
        self.plotDifferencing = plot_differencing
        self.plotSignalCorrelations = plot_signal_correlations
        self.plotFeatureImportance = plot_feature_importance
        self.plotScatterPlots = plot_scatter_plots

        # Register data
        self.data = data.astype('float64').fillna(0)
        if y is not None:
            if type(y) == pd.DataFrame:
                y = y[y.keys()[0]]
            if type(y) != pd.Series:
                y = pd.Series(y)
            self.Y = y.astype('float64').fillna(0)

        # General settings
        self.seasonPeriods = season_periods
        self.maxSamples = max_samples       # Time series
        self.differ = differ                # Correlations
        self.lags = lags                    # Correlations

        # Storage settings
        self.mode = mode
        self.tag = pre_tag
        self.version = version if version is not None else 0
        self.skip = skip_completed
        self.folder = folder if folder == '' or folder[-1] == '/' else folder + '/'
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

        # Create Base folder
        if not os.path.exists(self.folder + 'EDA/'):
            os.mkdir(self.folder + 'EDA')
        self.folder += 'EDA/'

    def run(self):
        # Run all functions
        print('[AutoML] Generating Missing Values Plot')
        self.missing_values()
        print('[AutoML] Generating Time plots')
        self.time_plots()
        print('[AutoML] Generating Box plots')
        self.box_plots()
        print('[AutoML] Generating Co-linearity Plots')
        self.co_linearity()

        # Functions which require output
        if self.Y is not None:
            print('[AutoML] Generating SHAP plot')
            self.shap()
            print('[AutoML] Generating Feature Ranking Plot')
            self.feature_ranking()

        # Mode specific functions
        if self.mode == 'classification':
            self._run_classification()
        else:
            self._run_regression()

    def _run_classification(self):
        pass

    def _run_regression(self):
        if self.plotSeasonality:
            print('[AutoML] Generating Seasonality Plots')
            self.seasonality()
        print('[AutoML] Generating Diff Var Plot')
        self.differencing()
        print('[AutoML] Generating ACF Plots')
        self.complete_auto_corr()
        print('[AutoML] Generating PACF Plots')
        self.partial_auto_corr()
        if self.Y is not None:
            print('[AutoML] Generating CCF Plots')
            self.cross_corr()
            print('[AutoML] Generating Scatter plots')
            self.scatters()

    def missing_values(self):
        if self.plotMissingValues:
            # Create folder
            if not os.path.exists(self.folder + 'MissingValues/'):
                os.mkdir(self.folder + 'MissingValues/')

            # Skip if exists
            if self.tag + 'v{}.png'.format(self.version) in os.listdir(self.folder + 'MissingValues/'):
                return

            # Plot
            fig, (a0, a1) = plt.subplots(1, 2, gridspec_kw={'width_ratios': [9, 1]})
            a0.imshow(self.data.isna(), cmap='Greys', aspect='auto', interpolation='none')
            a0.set_ylabel('Rows')
            a0.set_xticks(range(len(self.data.keys())), self.data.keys())
            a1.plot(self.data.isna().sum(axis=1), range(len(self.data)), c='k')
            a1.axis('off')
            fig.savefig(self.folder + 'MissingValues/v{}.png'.format(self.version))

    def box_plots(self):
        if self.plotBoxPlots:
            # Create folder
            if not os.path.exists(self.folder + 'BoxPlots/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'BoxPlots/v{}/'.format(self.version))

            # Iterate through vars
            for key in tqdm(self.data.keys()):

                # Skip if existing
                if self.tag + key + '.png' in os.listdir(self.folder + 'BoxPlots/v{}/'.format(self.version)):
                    continue

                # Figure prep
                fig = plt.figure(figsize=[24, 16])
                plt.title(key)

                # Classification
                if self.mode == 'classification':
                    plt.boxplot([self.data.loc[self.Y == 1, key], self.data.loc[self.Y == -1, key]], labels=['Faulty',
                                                                                                             'Healthy'])
                    plt.legend(['Faulty', 'Healthy'])

                # Regression
                if self.mode == 'regression':
                    plt.boxplot(self.data[key])

                # Store & Close
                fig.savefig(self.folder + 'BoxPlots/v{}/'.format(self.version) + self.tag + key + '.png',
                            format='png', dpi=300)
                plt.close()

    def time_plots(self):
        if self.plotTimePlots:
            # Create folder
            if not os.path.exists(self.folder + 'TimePlots/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'TimePlots/v{}/'.format(self.version))

            # Set pyplot limit
            matplotlib.use('Agg')
            matplotlib.rcParams['agg.path.chunksize'] = 200000

            # Under sample
            ind = np.linspace(0, len(self.data) - 1, self.maxSamples).astype('int')
            data, y = self.data.iloc[ind], self.Y.iloc[ind]

            # Iterate through features
            for key in tqdm(data.keys()):
                # Skip if existing
                if self.tag + key + '.png' in os.listdir(self.folder + 'TimePlots/v{}/'.format(self.version)):
                    continue

                # Figure preparation
                fig = plt.figure(figsize=[24, 16])
                plt.title(key)

                # Plot
                if self.mode == 'classification':
                    cm = plt.get_cmap('bwr')
                else:
                    cm = plt.get_cmap('summer')
                nm_output = (y - y.min()) / (y.max() - y.min())
                plt.scatter(data.index, data[key], c=cm(nm_output), alpha=0.3)

                # Store & Close
                fig.savefig(self.folder + 'TimePlots/v{}/'.format(self.version) + self.tag + key + '.png',
                            format='png', dpi=100)
                plt.close(fig)

    def seasonality(self):
        # Create folder
        if not os.path.exists(self.folder + 'Seasonality/'):
            os.mkdir(self.folder + 'Seasonality/')

        # Iterate through features
        for key in tqdm(self.data.keys()):
            if self.tag + key + '_v{}.png'.format(self.version) in os.listdir(self.folder + 'Seasonality/'):
                continue
            seasonality = STL(self.data[key], period=self.seasonPeriods).fit()
            fig = plt.figure(figsize=[24, 16])
            plt.plot(range(len(self.data)), self.data[key])
            plt.plot(range(len(self.data)), seasonality)
            plt.title(key + ', period=' + str(self.seasonPeriods))
            fig.savefig(self.folder + 'Seasonality/' + self.tag + str(self.seasonPeriods)+'/'+key +
                        '_v{}.png'.format(self.version), format='png', dpi=300)
            plt.close()

    def co_linearity(self):
        if self.plotCoLinearity:
            # Create folder
            if not os.path.exists(self.folder + 'CoLinearity/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'CoLinearity/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'MinimumRepresentation.png' in \
                    os.listdir(self.folder + 'CoLinearity/v{}/'.format(self.version)):
                return

            # Plot threshold matrix
            threshold = 0.95
            fig = plt.figure(figsize=[24, 16])
            plt.title('Co-linearity matrix, threshold {:.2f}'.format(threshold))
            plt.imshow(abs(self.data.corr()) < threshold, cmap='gray')
            fig.savefig(self.folder + 'CoLinearity/v{}/'.format(self.version) + self.tag + 'Matrix.png',
                        format='png', dpi=300)

            # Minimum representation
            corr_mat = self.data.corr().abs()
            upper = corr_mat.where(np.triu(np.ones(corr_mat.shape), k=1).astype('bool'))
            col_drop = [column for column in upper.columns if any(upper[column] > 0.95)]
            minimal_rep = self.data.drop(self.data[col_drop], axis=1)
            fig = plt.figure(figsize=[24, 16])
            plt.imshow(abs(minimal_rep.corr()) < threshold, cmap='gray')
            fig.savefig(self.folder + 'CoLinearity/v{}/'.format(self.version) + self.tag + 'Minimum_Representation.png',
                        format='png', dpi=300)

    def differencing(self):
        if self.plotDifferencing:
            # Create folder
            if not os.path.exists(self.folder + 'Lags/'):
                os.mkdir(self.folder + 'Lags/')

            # Skip if existing
            if self.tag + 'Variance.png' in os.listdir(self.folder + 'Lags/'):
                return

            # Setup
            max_lags = 4
            var_vec = np.zeros((max_lags, len(self.data.keys())))
            diff_data = self.data / np.sqrt(self.data.var())

            # Calculate variance per lag
            for i in range(max_lags):
                var_vec[i, :] = diff_data.var()
                diff_data = diff_data.diff(1)[1:]

            # Plot
            fig = plt.figure(figsize=[24, 16])
            plt.title('Variance for different lags')
            plt.plot(var_vec)
            plt.xlabel('Lag')
            plt.yscale('log')
            plt.ylabel('Average variance')
            fig.savefig(self.folder + 'Lags/' + self.tag + 'Variance.png', format='png', dpi=300)

    def complete_auto_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/ACF/'):
                os.makedirs(self.folder + 'Correlation/ACF/')

            # Difference data
            diff_data = copy.copy(self.data)
            for i in range(self.differ):
                diff_data = diff_data.diff(1)[1:]

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v{}.png'.format(self.version) in \
                        os.listdir(self.folder + 'Correlation/ACF/'):
                    continue

                # Plot
                fig = plot_acf(diff_data[key], fft=True)
                plt.title(key)
                fig.savefig(self.folder + 'Correlation/ACF/' + self.tag + key + '_differ_' + str(self.differ) +
                            '_v{}.png'.format(self.version), format='png', dpi=300)
                plt.close()

    def partial_auto_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/PACF/'):
                os.makedirs(self.folder + 'Correlation/PACF/')

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v{}.png'.format(self.version) in \
                        os.listdir(self.folder + 'EDA/Correlation/PACF/'):
                    continue

                # Plot
                try:
                    fig = plot_pacf(self.data[key])
                    fig.savefig(self.folder + 'EDA/Correlation/PACF/' + self.tag + key + '_differ_' +
                                str(self.differ) + '_v{}.png'.format(self.version), format='png', dpi=300)
                    plt.title(key)
                    plt.close()
                except Exception as e:
                    # todo find exception
                    raise ValueError(e)

    def cross_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/Cross/'):
                os.makedirs(self.folder + 'Correlation/Cross/')

            # Prepare
            folder = 'Correlation/Cross/'
            y = self.Y.to_numpy().reshape((-1))

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v{}.png'.format(self.version) in \
                        os.listdir(self.folder + folder):
                    continue

                # Plot
                try:
                    fig = plt.figure(figsize=[24, 16])
                    plt.xcorr(self.data[key], y, maxlags=self.lags)
                    plt.title(key)
                    fig.savefig(self.folder + folder + self.tag + key + '_differ_' + str(self.differ) +
                                '_v{}.png'.format(self.version), format='png', dpi=300)
                    plt.close()
                except Exception as e:
                    raise ValueError(e)

    def scatters(self):
        if self.plotScatterPlots:
            # Create folder
            if not os.path.exists(self.folder + 'Scatters/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'Scatters/v{}/'.format(self.version))

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if '{}{}.png'.format(self.tag, key) in os.listdir(self.folder + 'Scatters/v{}/'.format(self.version)):
                    continue

                # Plot
                fig = plt.figure(figsize=[24, 16])
                plt.scatter(self.Y, self.data[key], alpha=0.2)
                plt.ylabel(key)
                plt.xlabel('Output')
                plt.title('Scatter Plot ' + key + ' - Output')
                fig.savefig(self.folder + 'Scatters/v{}/'.format(self.version) + self.tag + key + '.png',
                            format='png', dpi=100)
                plt.close(fig)

    def shap(self, args=None):
        args = args if args is not None else {}
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v{}/'.format(self.version)):
                os.makedirs(self.folder + 'Features/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'SHAP.png' in os.listdir(self.folder + 'Features/v{}/'.format(self.version)):
                return

            # Create model
            if self.mode == 'classification':
                model = RandomForestClassifier(**args).fit(self.data, self.Y)
            else:
                model = RandomForestRegressor(**args).fit(self.data, self.Y)

            # Calculate SHAP values
            import shap
            shap_values = shap.TreeExplainer(model).shap_values(self.data)

            # Plot
            fig = plt.figure(figsize=[8, 32])
            plt.subplots_adjust(left=0.4)
            shap.summary_plot(shap_values, self.data, plot_type='bar')
            fig.savefig(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'SHAP.png', format='png', dpi=300)

    def feature_ranking(self, **args):
        args = args if args is not None else {}
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v{}/'.format(self.version)):
                os.mkdir(self.folder + 'Features/v{}/'.format(self.version))

            # Skip if existing
            if self.tag + 'RF.png' in os.listdir(self.folder + 'Features/v{}/'.format(self.version)):
                return

            # Create model
            if self.mode == 'classification':
                model = RandomForestClassifier(**args).fit(self.data, self.Y)
            else:
                model = RandomForestRegressor(**args).fit(self.data, self.Y)

            # Plot
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ind = np.argsort(model.feature_importances_)
            plt.barh(list(self.data.keys()[ind])[-15:], width=model.feature_importances_[ind][-15:],
                     color='#2369ec')
            fig.savefig(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'RF.png', format='png', dpi=300)
            plt.close()

            # Store results
            results = pd.DataFrame({'x': self.data.keys(), 'score': model.feature_importances_})
            results.to_csv(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'RF.csv')

    # def predictive_power_score(self):
    #     if self.plotFeatureImportance:
    #         # Create folder
    #         if not os.path.exists(self.folder + 'Features/v{}/'.format(self.version)):
    #             os.mkdir(self.folder + 'Features/v{}/'.format(self.version))
    #
    #         # Skip if existing
    #         if self.tag + 'PPScore.png' in os.listdir(self.folder + 'Features/v{}/'.format(self.version)):
    #             return
    #
    #         # Calculate PPS
    #         data = self.data.copy()
    #         if isinstance(self.Y, pd.core.series.Series):
    #             data.loc[:, 'target'] = self.Y
    #         elif isinstance(self.Y, pd.DataFrame):
    #             data.loc[:, 'target'] = self.Y.loc[:, self.Y.keys()[0]]
    #         pp_score = ppscore.predictors(data, 'target').sort_values('ppscore')
    #
    #         # Plot
    #         fig, ax = plt.subplots(figsize=[4, 6])
    #         plt.subplots_adjust(left=0.5, top=1, bottom=0)
    #         ax.spines['right'].set_visible(False)
    #         ax.spines['bottom'].set_visible(False)
    #         ax.spines['top'].set_visible(False)
    #         plt.barh(pp_score['x'][-15:], width=pp_score['ppscore'][-15:], color='#2369ec')
    #         fig.savefig(self.folder + 'Features/v{}/'.format(self.version) + self.tag + 'Ppscore.png', format='png',
    #                     dpi=400)
    #         plt.close()
    #
    #         # Store results
    #         pp_score.to_csv(self.folder + 'Features/v{}/pp_score.csv'.format(self.version))
